body {
    margin: auto;
    padding-right: 1em;
    padding-left: 1em;
    margin-left: 2em;
    border-left: 1px solid black;
    color: black;
    font-size: 100%;
    line-height: 140%;
    color: #333;
}

pre {
    border: 1px dotted gray;
    background-color: #ececec;
    color: #1111111;
    padding: 0.5em;
    line-height: 1.42857143;
    tab-size: 4;
    -moz-tab-size: 4;
}

code {
    font-family: monospace;
}

h1 a, h2 a, h3 a, h4 a, h5 a {
    text-decoration: none;
    color: #7a5ada;
}
h1, h2, h3, h4, h5 {
    font-family: sans-serif;
    font-weight: bold;
    text-decoration: underline;
    color: #7a5ada;
}
h1 {
    font-size: 130%;
}
h2 {
    font-size: 110%;
}
h3 {
    font-size: 95%;
}
h4 {
    font-size: 90%;
    font-style: italic;
}
h5 {
    font-size: 90%;
    font-style: italic;
}
h1.title {
    font-size: 200%;
    font-weight: bold;
    padding-top: 0.2em;
    padding-bottom: 0.2em;
    text-align: left;
    border: none;
}

dt code {
    font-weight: bold;
}
dd p {
    margin-top: 0;
}

#TOC {
    float: right;
    width: 40%;
    background: #eee;
    font-size: 0.8em;
    padding: 1em 2em;
    margin: 0.0 0.5em 0.5em;
}
